import numpy as np
import sys

from .core import *

__all__ = [
    "add",
    "subtract",
    "multiply",
    "divide",
    "matmul",
    "pow",
    "square",
    "cube",
    "sin",
    "cos",
    "log",
    "exp",
    "reduce_mean",
    "reduce_sum",
    "sigmoid",
]


def _convert_to_tensor(item):
    if not isinstance(item, Tensor):
        return constant(item)
    return item


# Binary operations
def _binary_operation(a, b, op):
    a = _convert_to_tensor(a)
    b = _convert_to_tensor(b)
    return Tensor(operation=op, inputs=[a, b], dtype=a.dtype, shape=a.shape)


def add(a, b):
    op = sys._getframe().f_code.co_name
    return _binary_operation(a, b, op)


def subtract(a, b):
    op = sys._getframe().f_code.co_name
    return _binary_operation(a, b, op)


def multiply(a, b):
    op = sys._getframe().f_code.co_name
    return _binary_operation(a, b, op)


def divide(a, b):
    op = sys._getframe().f_code.co_name
    return _binary_operation(a, b, op)


def pow(a, b):
    op = sys._getframe().f_code.co_name
    return _binary_operation(a, b, op)


def square(a):
    return pow(a, 2)


def cube(a):
    return pow(a, 3)


def matmul(a, b):
    a = _convert_to_tensor(a)
    b = _convert_to_tensor(b)
    if (len(a.shape) < 2) or (len(b.shape) < 2):
        error = f"Both the arguments should be 2 dimensional arrays, dimensions passed {a.shape}, {b.shape}."
        raise ValueError(error)
    if a.shape[1] != b.shape[0]:
        error = f"Array dimensions not compatible for matrix multiplication, dimensions passed {a.shape}, {b.shape}."
        raise ValueError(error)

    op = sys._getframe().f_code.co_name
    return Tensor(
        operation=op, inputs=[a, b], dtype=a.dtype, shape=(a.shape[0], b.shape[1])
    )


# Unary operations
def _unary_operation(a, op):
    a = _convert_to_tensor(a)
    return Tensor(operation=op, inputs=[a], dtype=a.dtype, shape=a.shape)


def sin(a):
    op = sys._getframe().f_code.co_name
    return _unary_operation(a, op)


def cos(a):
    op = sys._getframe().f_code.co_name
    return _unary_operation(a, op)


def log(a):
    op = sys._getframe().f_code.co_name
    return _unary_operation(a, op)


def exp(a):
    op = sys._getframe().f_code.co_name
    return _unary_operation(a, op)


def reduce_mean(a):
    op = sys._getframe().f_code.co_name
    return _unary_operation(a, op)


def reduce_sum(a):
    op = sys._getframe().f_code.co_name
    return _unary_operation(a, op)


def sigmoid(a):
    op = sys._getframe().f_code.co_name
    return _unary_operation(a, op)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yodf']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.4,<2.0.0']

setup_kwargs = {
    'name': 'yodf',
    'version': '1.0.13',
    'description': "'Hello, World!' Forward Mode Autdiff library with Tensorflow 1 like interface.",
    'long_description': '## A \'Hello, World!\' forward mode autodiff library.\n\nThis small (~500 lines) library is meant as an illustration of how forward mode autodiff can possibly be implemented. It lets you compute the value and the derivative of a function expressed as a computational flow using the primitives provided by the library. Interface of the library is very similar to Tensorflow 1.\\*. All the samples provided in _examples_ folder can very well be run if you do **import tensorflow as tf** as opposed to **import yodf as tf** It supports following operations { "add", "subtract", "divide", "multiply", "pow", "sin", "cos", "log", "exp", "matmul", "sigmoid", "reduce_mean", "reduce_sum" }.\n\n### Installation\n\n**pip install yodf** will install the library. Only dependency it has is _numpy_. Samples provided in examples folder also have dependency on _matplotlib_ and _scipy_.\n\n### Basic usage\n\nBelow code computes the value and the derivative of the function **x^2** at x=5.0\n\n```\nimport yodf as tf\nx = tf.Variable(5.0)\ncost = x**2\nwith tf.Session() as s:\n    # global_variables_initializer API added just so as to\n    # resemble Tensorflow, it hardly does anything\n    s.run(tf.global_variables_initializer())\n    s.run(cost)\nprint(x.value, cost.value, cost.gradient)\n\n## Output\n## 5.0 25.0 10.0\n```\n\n### Basic gradient descent example\n\nBelow code computes optima of the function **x^2** along with the value at which optima occurs starting with x=5.0\n\n```\nimport yodf as tf\nx = tf.Variable(5.0)\ncost = x**2\ntrain = tf.train.GradientDescentOptimizer(learning_rate=0.2).minimize(cost)\nwith tf.Session() as s:\n    s.run(tf.global_variables_initializer())\n    for _ in range(50):\n        _, cost_final, x_final = s.run([train, x, cost])\nprint(f"Minima: {cost_final:.10f}, x at minima: {x_final:.10f}")\n\n## Output\n## Minima: 0.0000000000, x at minima: 0.0000000000\n```\n\n## How does it work?\n\nIt has a class called _Tensor_ with _Variable_ and _\\_Constant_ as subclasses. Tensor object holds a value and a gradient. Gradient of a constant is 0 and that of a variable is 1 which is as good as saying d(x)/dx.  \nA tensor can also represent an operation and a tensor representating an operation gets created using a convenient function call like _tf.sin()_ or _tf.matmul()_ etc.\n\n```\nimport numpy as np\nimport yodf as tf\nx = tf.Variable(np.array([[1,1],[2,2]]))\nop_sin = tf.sin(x)\nprint(op_sin)\n\n## Output\n## <yod.Tensor type=TensorType.INT, shape=(2, 2), operation=\'sin\'>\n```\n\nYou typically pass a tensor to run method of _Session_ class which ends up evaluating the tensor along with its derivative. Execute method of tensor just knows how to compute derivative of basic arithmatic operations, power function and some of the transcendental functions like sin, cos, log, exp. It also knows how to compute derivative when matrix multiplication operation is involved. By applying the chain rule repeatedly to these operations, derivative of an arbitrary function (represented as a tensor) gets computed automatically. _run_ method simply builds post order traversal tree of the tensor passed to it and evaluates all the nodes in the tree. _GradientDescentOptimizer_ simply updates the value of the variable based on the gradient of the cost tensor passed to its minimize function.  \nWith multiple independent variables, partial derivative of one variable gets computed at a time while the gradient of rest of the variables is set to 0. This, in turn, is done for all the variables and partial derivatives or the gradients of all the vatiables are accumulated by _GradientDescentOptimizer_ which is not necessarily very clean.\n\n## Examples\n\nExamples folder shows use of this library for\n\n1. <a href="https://github.com/yogimogi/yodf/blob/master/examples/example1_simple_cost_function.ipynb">A gradient descent problem for a simple cost function</a>\n2. <a href="https://github.com/yogimogi/yodf/blob/master/examples/example2_cost_function_2_variables.ipynb">A gradient descent problem for a simple cost function with 2 independent variables</a>\n3. <a href="https://github.com/yogimogi/yodf/blob/master/examples/example3_linear_regression.ipynb">A linear regression problem</a>\n4. <a href="https://github.com/yogimogi/yodf/blob/master/examples/example4_logistic_regression.ipynb">A logistic regression problem</a>\n5. <a href="https://github.com/yogimogi/yodf/blob/master/examples/example5_neural_network.ipynb">A neural network with one hidden layer and one output</a>\n6. <a href="https://github.com/yogimogi/yodf/blob/master/examples/example6_neural_network_mnist.ipynb">A neural network with one hidden layer and 10 outputs (MNIST digit classification)</a>\n\n## Limitiation of forward mode autodiff\n\nThough with forward mode autodiff, derivative of a function with one independent variables gets computed during forward pass itself and no backward pass is needed as is the case with reverse mode autodiff (generalized backpropagation), with multiple indepdent variables (say weights in a neural network), as many passes are needed as number of indepdent variables. So as can be seen in <a href="https://github.com/yogimogi/yodf/blob/master/examples/example3_linear_regression.ipynb">linear regression sample</a>, time needed by gradient descent linearly increases with increase in degree of polynomial you are trying to fit. For MNIST digit classification, this library becomes almost unusable due to large number of independent variables whose gradient needs to be computed.\n',
    'author': 'Yogesh Ketkar',
    'author_email': 'yogimogi@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/yogimogi/yodf',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from enum import Enum
import numpy as np
from .train import GradientDescentOptimizer

"""
    __init__.py of yod module does
    'from .core import *'
    Only members listed below will be accessbiel through yod when
    yod is imported as 'import yod'
    If you gets imported as 'from yod import *', then members from
    the list will go in the global import namespace which is not nice
"""
__all__ = [
    "TensorType",
    "Tensor",
    "Variable",
    "constant",
    "placeholder",
    "Session",
    "global_variables_initializer",
]


class TensorType(Enum):
    INT = 1
    FLOAT = 2


class Tensor:
    SUPPORTED_OPERATIONS = {
        # binary operations
        "add",
        "subtract",
        "divide",
        "multiply",
        "pow",
        # transcendental functions
        "sin",
        "cos",
        "reduce_mean",
        "reduce_sum",
        "log",
        "exp",
        "matmul",
        "sigmoid",
    }

    def __add__(self, other):
        from .operations import add

        return add(self, other)

    def __radd__(self, other):
        from .operations import add

        return add(other, self)

    def __sub__(self, other):
        from .operations import subtract

        return subtract(self, other)

    def __rsub__(self, other):
        from .operations import subtract

        return subtract(other, self)

    def __neg__(self):
        from .operations import multiply

        return multiply(self, -1)

    def __mul__(self, other):
        from .operations import multiply

        return multiply(self, other)

    def __rmul__(self, other):
        from .operations import multiply

        return multiply(other, self)

    def __truediv__(self, other):
        from .operations import divide

        return divide(self, other)

    def __rtruediv__(self, other):
        from .operations import divide

        return divide(other, self)

    def __pow__(self, other):
        from .operations import pow

        return pow(self, other)

    def __init__(self, dtype=None, value=None, shape=None, operation=None, inputs=None):
        if shape is not None:
            self.shape = shape
        else:
            self.shape = ()
        self.dtype = dtype
        self.value = None
        # 'if value' will not work with numpy arrays if it is not None
        if value is not None:
            self.__validate_and_set_value(value)

        if operation is not None and operation not in Tensor.SUPPORTED_OPERATIONS:
            error = f"Operation {operation} not supported. Must be one of {Tensor.SUPPORTED_OPERATIONS}."
            raise ValueError(error)
        self.operation = operation
        self.inputs = inputs

        self.gradient = None

    def __validate_and_set_value(self, value):
        if isinstance(value, int):
            self.dtype = TensorType.INT
        elif isinstance(value, float):
            self.dtype = TensorType.FLOAT
        elif isinstance(value, np.ndarray):
            if value.ndim != 2:
                raise ValueError(f"Only 2 dimensional Numpy arrays supported.")
            np_dtype = str(value.dtype)
            if np_dtype.startswith("int") or np_dtype.startswith("uint"):
                self.dtype = TensorType.INT
            elif np_dtype.startswith("float"):
                self.dtype = TensorType.FLOAT
            else:
                error = f"Numpy array with unsupported data type {np_dtype}."
                raise ValueError(error)
            self.shape = value.shape
        else:
            error = f"Unsupported Tensor Type {type(value)}."
            raise ValueError(error)
        self.value = value

    def set_gradient(self, g):
        if self.shape == ():
            self.gradient = g
        else:
            self.gradient = np.ones(self.shape) * g

    def __str__(self):
        if self.operation is None:
            return f"<yod.Tensor type={str(self.dtype)}, shape={self.shape}>"
        else:
            return f"<yod.Tensor type={str(self.dtype)}, shape={self.shape}, operation='{self.operation}'>"

    def __sigmoid(self, x):
        return 1 / (1 + np.e ** -x)

    def __reduce(self, operation):
        if operation == "sum":
            self.value = np.sum(self.inputs[0].value)
            self.gradient = np.sum(self.inputs[0].gradient)
        elif operation == "mean":
            self.value = np.mean(self.inputs[0].value)
            self.gradient = np.mean(self.inputs[0].gradient)
        else:
            error = f"Unsupported operation for reduction {operation}."
            raise ValueError(error)

    def execute(self, feed_dict=None):
        if self.operation == "add":
            self.value = self.inputs[0].value + self.inputs[1].value
            self.gradient = self.inputs[0].gradient + self.inputs[1].gradient
        elif self.operation == "subtract":
            self.value = self.inputs[0].value - self.inputs[1].value
            self.gradient = self.inputs[0].gradient - self.inputs[1].gradient
        elif self.operation == "multiply":
            self.value = self.inputs[0].value * self.inputs[1].value
            self.gradient = (
                self.inputs[0].value * self.inputs[1].gradient
                + self.inputs[0].gradient * self.inputs[1].value
            )
        elif self.operation == "divide":
            self.value = self.inputs[0].value / self.inputs[1].value
            self.gradient = (
                self.inputs[0].gradient * self.inputs[1].value
                - self.inputs[0].value * self.inputs[1].gradient
            ) / (self.inputs[1].value ** 2)
        elif self.operation == "pow":
            raise_to = self.inputs[1].value
            self.value = np.power(self.inputs[0].value, raise_to)
            self.gradient = (
                raise_to
                * np.power(self.inputs[0].value, raise_to - 1)
                * self.inputs[0].gradient
            )
        elif self.operation == "sin":
            self.value = np.sin(self.inputs[0].value)
            self.gradient = np.cos(self.inputs[0].value) * self.inputs[0].gradient
        elif self.operation == "cos":
            self.value = np.cos(self.inputs[0].value)
            self.gradient = -np.sin(self.inputs[0].value) * self.inputs[0].gradient
        elif self.operation == "log":
            self.value = np.log(self.inputs[0].value)
            self.gradient = self.inputs[0].gradient / self.inputs[0].value
        elif self.operation == "exp":
            self.value = np.e ** self.inputs[0].value
            self.gradient = self.value * self.inputs[0].gradient
        elif self.operation == "matmul":
            self.value = np.dot(self.inputs[0].value, self.inputs[1].value)
            self.gradient = np.dot(
                self.inputs[0].value, self.inputs[1].gradient
            ) + np.dot(self.inputs[0].gradient, self.inputs[1].value)
        elif self.operation == "reduce_mean":
            self.__reduce("mean")
        elif self.operation == "reduce_sum":
            self.__reduce("sum")
        elif self.operation == "sigmoid":
            self.value = self.__sigmoid(self.inputs[0].value)
            self.gradient = self.value * (1 - self.value) * self.inputs[0].gradient


class _Constant(Tensor):
    def __init__(self, value):
        super().__init__(value=value)
        self.set_gradient(0)


def constant(value):
    return _Constant(value=value)


class _PlaceHolder(Tensor):
    def __init__(self, dtype, shape=None):
        super().__init__(dtype=dtype, shape=shape)
        self.value_fed = None

    def execute(self, feed_dict=None):
        if feed_dict is None:
            raise Exception("Feed dictionary missing.")
        if feed_dict is None or self not in feed_dict:
            raise Exception("Feed dictionary missing required placeholder.")

        if self.value_fed is feed_dict[self]:
            return
        self.value_fed = feed_dict[self]

        from .operations import _convert_to_tensor

        t = _convert_to_tensor(feed_dict[self])
        self.value = t.value
        self.gradient = t.gradient


def placeholder(dtype, shape):
    return _PlaceHolder(dtype=dtype, shape=shape)


class Variable(Tensor):
    def __init__(self, initial_value):
        super().__init__(value=initial_value)
        self.set_gradient(1)
        self.__initialized = False
        _graph_completely_useless_.add_variable(self)

    def __str__(self):
        return f"<yod.Variable type={str(self.dtype)}, shape={self.shape}>"

    def initialize(self):
        self.__initialized = True

    def execute(self, feed_dict=None):
        if not self.__initialized:
            error = f"Attempt to use uninitialized variable {self}."
            raise Exception(error)


class Session:
    def __init__(self):
        global _session_yodf_
        _session_yodf_ = self
        self.tensors = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        global _session
        _session_yodf_ = None
        self.tensors = {}

    def __post_order(self, t):
        tensors_po = []

        def _append(t):
            for ti in tensors_po:
                if t is ti:
                    return
            tensors_po.append(t)

        def _traverse(t):
            if t.operation is not None:
                for ti in t.inputs:
                    _traverse(ti)
            _append(t)

        _traverse(t)
        return tensors_po

    def get_variables(self, t):
        tensors_po = self.__post_order(t)
        return list(filter(lambda t: isinstance(t, Variable), tensors_po))

    def __run(self, t, feed_dict=None):
        if isinstance(t, InitOperation):
            t.init_variables()
            return

        if isinstance(t, GradientDescentOptimizer):
            t.execute(feed_dict)
            return

        if not isinstance(t, Tensor):
            error = f"t <{t}> is not an instance of yod.Tensor."
            raise ValueError(error)

        if t not in self.tensors:
            tensors_po = self.__post_order(t)
            self.tensors[t] = tensors_po
        else:
            tensors_po = self.tensors[t]

        for tp in tensors_po:
            tp.execute(feed_dict)

        return t.value

    def run(self, t, feed_dict=None):
        if isinstance(t, list):
            values = []
            for ti in t:
                values.append(self.__run(ti, feed_dict))
            return values
        else:
            return self.__run(t, feed_dict)


# Class absolutely not needed or used for the working of the package
# Created just to be able to add something to global_variables_initializer
class Graph:
    def __init__(self):
        self.__variables = []

    def cleanup(self):
        self.__variables = []

    def add_variable(self, v):
        self.__variables.append(v)

    def get_variables(self):
        return self.__variables

    def init_variables(self):
        for v in self.__variables:
            v.initialize()


global _graph_completely_useless_
_graph_completely_useless_ = Graph()

global _session_yodf_
_session_yodf_ = None


def get_session():
    return _session_yodf_


class InitOperation:
    def __init__(self):
        self.value = None

    def init_variables(self):
        _graph_completely_useless_.init_variables()


def global_variables_initializer():
    return InitOperation()



import numpy as np


class GradientDescentOptimizer:
    # Minimize supports only the operations listed below
    # if Tensor passed to it is not a literal
    MINIMZE_SUPPORTED_OPERATIONS = {
        'reduce_mean',
        'reduce_sum',
    }

    def __init__(self, learning_rate):
        self.learning_rate = learning_rate
        self.cost = None

    def minimize(self, cost):
        if len(cost.shape) == 0 or cost.operation in GradientDescentOptimizer.MINIMZE_SUPPORTED_OPERATIONS:
            self.cost = cost
        else:
            error = f"Neither Tensor value is a literal nor is the operation " + \
                f"one from {GradientDescentOptimizer.SUPPORTED_OPERATIONS_MINIMZE}."
            raise ValueError(error)
        return self

    # Variable (basically a Tensor) can either be a literal
    # or 2 dimensional numpy array. This function handles
    # latter type of variables
    def __execute_variable_matrix(self, v, feed_dict=None):
        from .core import get_session
        s = get_session()

        gradient = np.zeros(v.shape)
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v.gradient.fill(0)
                v.gradient[i][j] = 1
                s.run(self.cost, feed_dict)
                gradient[i][j] = self.cost.gradient
        return gradient

    def execute(self, feed_dict=None):
        from .core import get_session
        s = get_session()

        vars = s.get_variables(self.cost)
        gradients = []
        for i in range(len(vars)):
            vars[i].set_gradient(1)
            for j in range(len(vars)):
                if i != j:
                    vars[j].set_gradient(0)
            if len(vars[i].shape) > 0:
                gradients.append(
                    self.__execute_variable_matrix(vars[i], feed_dict))
            else:
                s.run(self.cost, feed_dict)
                gradients.append(self.cost.gradient)

        # Restoring gradients of variables to 1
        for i in range(len(vars)):
            vars[i].set_gradient(1)

        # Need to update all variables only after all the gradients have been computed.
        # Updating a variable immediately after its gradient is computed would result
        # in subsqeunt gradients being wrongly computed
        for i in range(len(gradients)):
            vars[i].value = vars[i].value-self.learning_rate*gradients[i]
